------------------------------------------------------------------------------
--                                                                          --
--                        GNAT RUN-TIME COMPONENTS                          --
--                                                                          --
--                          S Y S T E M . C R T L                           --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--          Copyright (C) 2003-2026, Free Software Foundation, Inc.         --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 3,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.                                     --
--                                                                          --
-- As a special exception under Section 7 of GPL version 3, you are granted --
-- additional permissions described in the GCC Runtime Library Exception,   --
-- version 3.1, as published by the Free Software Foundation.               --
--                                                                          --
-- You should have received a copy of the GNU General Public License and    --
-- a copy of the GCC Runtime Library Exception along with this program;     --
-- see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see    --
-- <http://www.gnu.org/licenses/>.                                          --
--                                                                          --
-- GNAT was originally developed  by the GNAT team at  New York University. --
-- Extensive contributions were provided by Ada Core Technologies Inc.      --
--                                                                          --
------------------------------------------------------------------------------

package body System.CRTL is

   ----------
   -- read --
   ----------

   function read (fd : int; buffer : chars; count : size_t) return ssize_t
   is
      function read_raw
        (fd : int; buffer : chars; count : size_t) return ssize_t;
      pragma Import (C, read_raw, "read");
   begin
      return read_raw (fd, buffer, count);
   end read;

   -----------
   -- write --
   -----------

   function write (fd : int; buffer : chars; count : size_t) return ssize_t
   is
      function write_raw
        (fd : int; buffer : chars; count : size_t) return ssize_t;
      pragma Import (C, write_raw, "write");
   begin
      return write_raw (fd, buffer, count);
   end write;

end System.CRTL;
